#ifndef RT_MEM_ARENA_H
#define RT_MEM_ARENA_H

#include "runtime.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef struct rt_arena_s {
    void *base;
    size_t at;
    size_t size;
    int needs_free;
} rt_arena;

typedef struct {
    int ok;
    rt_arena arena;
} rt_create_arena_result;

typedef struct {
    size_t at;
} rt_arena_rewindpoint;

RT_DLLEXPORT rt_create_arena_result rtCreateArena(void *memory, size_t size);

RT_DLLEXPORT void rtReleaseArena(rt_arena *arena);

RT_DLLEXPORT void *rtArenaPush(rt_arena *arena, size_t n);

RT_DLLEXPORT void *rtArenaPushZero(rt_arena *arena, size_t n);

RT_DLLEXPORT void rtArenaPop(rt_arena *arena, size_t n);

RT_INLINE void rtArenaClear(rt_arena *arena) {
    arena->at = 0;
}

RT_INLINE rt_arena_rewindpoint rtGetArenaRewindPoint(rt_arena *arena) {
#ifndef __cplusplus
    return (rt_arena_rewindpoint){.at = arena->at};
#else
    rt_arena_rewindpoint rp = {arena->at};
    return rp;
#endif
}

RT_INLINE void rtArenaRewind(rt_arena *arena, rt_arena_rewindpoint rewind) {
#ifndef NDEBUG
    if (rewind.at > arena->at)
        rtReportError("CORE",
                      "Tried to rewind an arena to a point that lies beyond the last allocation.");
#endif
    arena->at = rewind.at;
}

/* Helper macros */
#define RT_ARENA_PUSH_STRUCT(_Arena, _Type)         rtArenaPush((_Arena), sizeof(_Type))
#define RT_ARENA_PUSH_STRUCT_ZERO(_Arena, _Type)    rtArenaPushZero((_Arena), sizeof(_Type))
#define RT_ARENA_POP_STRUCT(_Arena, _Type)          rtArenaPop((_Arena), sizeof(_Type))
#define RT_ARENA_PUSH_ARRAY(_Arena, _Type, _N)      rtArenaPush((_Arena), sizeof(_Type) * (_N))
#define RT_ARENA_PUSH_ARRAY_ZERO(_Arena, _Type, _N) rtArenaPushZero((_Arena), sizeof(_Type) * (_N))
#define RT_ARENA_POP_ARRAY(_Arena, _Type, _N)       rtArenaPop((_Arena), sizeof(_Type) * (_N)

#ifdef __cplusplus
}
#endif

#endif
