#ifndef RT_GFX_BACKEND_H
#define RT_GFX_BACKEND_H

/* Backend functions and types. */

#include <stddef.h>

#include "gfx.h"
#include "runtime.h"
#include "assets.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifdef _WIN32
struct HINSTANCE__;
struct HWND__;
#elif defined(RT_USE_XLIB)
struct _XDisplay;
#endif

struct rt_renderer_init_info_s {
#ifdef _WIN32
    struct HINSTANCE__ *hInstance;
    struct HWND__ *hWnd;
#elif defined(RT_USE_XLIB)
    struct _XDisplay *display;
    unsigned long window;
#endif
};

typedef struct {
    /* rt_uid vertex_shader;
    rt_uid fragment_shader;
    rt_uid compute_shader;
    */

    rt_relptr texture_bindings;
    rt_relptr uniform_bindings;
    rt_relptr storage_bindings;

    uint16_t texture_binding_count;
    uint16_t uniform_binding_count;
    uint16_t storage_binding_count;
} rt_pipeline_info;

typedef void rt_register_renderer_cvars_fn(void);
typedef rt_result rt_init_renderer_fn(const rt_renderer_init_info *info);
typedef void rt_shutdown_renderer_fn(void);
typedef rt_pipeline_handle rt_compile_pipeline_fn(const rt_pipeline_info *info);
typedef void rt_destroy_pipeline_fn(rt_pipeline_handle handle);

typedef struct {
    rt_register_renderer_cvars_fn *RegisterCVars;
    rt_init_renderer_fn *Init;
    rt_shutdown_renderer_fn *Shutdown;
    rt_compile_pipeline_fn *CompilePipeline;
    rt_destroy_pipeline_fn *DestroyPipeline;
} rt_renderer_api;

#define RT_RENDERER_API_FN(name) RT_DLLEXPORT rtRen##name

#ifndef RT_DONT_DEFINE_RENDERER_GLOBAL
extern rt_renderer_api g_renderer;
#endif

#ifdef __cplusplus
}
#endif

#endif
