#include "utils.h"
#include "assetmeta.h"
#include "processing.h"
#include "assetsettings.h"
#include "packages.h"

#include <string.h>

#include "runtime/uidtab.h"

typedef struct {
    char path_scratch[1024];
    unsigned int path_end;
} rt_discovery_data;

static rt_result LoadCompressedAsset(rt_uid uid, void **buffer, size_t size) {
}

static rt_result DirectoryHandler(const char *name, rtIterateDirElementType type, void *user) {
    rt_discovery_data *data = user;

    size_t name_len = strlen(name);

    if (type == RT_DIR_ELEMENT_TYPE_FILE) {
        /* Skip files we don't want to process */
        if (name_len >= 3) {
            if (memcmp(&name[name_len - 3], ".as", 3) == 0)
                return RT_SUCCESS;
        }
        if (name_len >= 4) {
            if (memcmp(&name[name_len - 4], ".pkg", 4) == 0)
                return RT_SUCCESS;
            else if (memcmp(&name[name_len - 4], ".bin", 4) == 0)
                return RT_SUCCESS;
        }
        if (strcmp(name, "packages.txt") == 0)
            return RT_SUCCESS;
        if (name[0] == '.') {
            return RT_SUCCESS;
        }

        /* Check if we know that file */
        if (data->path_end > 0) {
            data->path_scratch[data->path_end] = '/';
            memcpy(data->path_scratch + data->path_end + 1, name, name_len);
            data->path_scratch[data->path_end + 1 + name_len] = '\0';
        } else {
            memcpy(data->path_scratch, name, name_len);
            data->path_scratch[name_len] = '\0';
        }

        rt_file_id fid = rtAddFile(data->path_scratch);
        if (rtLookupUID(fid) != RT_INVALID_UID) {
            rt_assetmeta meta = {0};
            if (!rtGetAssetMeta(fid, &meta) || (meta.last_changed >= meta.compiled_ts)) {
                /* The file (may have) changed */
                rt_result res = rtAddFileToProcessingQueue(fid, meta.processing_flags);
                if (res != RT_SUCCESS)
                    return res;
            }
            else {
                /* The file is unchanged, we just need to add it to the output again. */
                rtLog("ASSETC", "File %s is unchanged.", data->path_scratch);

                rt_asset_settings settings = {0};
                if (rtLoadAssetSettings(data->path_scratch, &settings) != RT_SUCCESS) {
                    rtLog("ASSETC", "Failed to load settings for %s", data->path_scratch);
                    return RT_UNKNOWN_ERROR;
                }

                /* We need to load the processed data and add it to the package */
                rt_uid uid = rtLookupUID(fid);
                if (uid == RT_INVALID_UID) {
                    rtLog("ASSETC", "Failed to lookup UID of known asset %s", data->path_scratch);
                    return RT_UNKNOWN_ERROR;
                }
                rtAddUnchangedAssetToPackage(settings.package, uid);
            }
        } else {
            /* Process it */
            rt_asset_settings settings = {0};
            if (rtLoadAssetSettings(data->path_scratch, &settings) != RT_SUCCESS) {
                rtLog("ASSETC", "Failed to load settings for %s", data->path_scratch);
            }
            
            rt_result res = rtAddFileToProcessingQueue(fid, settings.processing_flags);
            if (res != RT_SUCCESS)
                return res;
        }
    } else if (type == RT_DIR_ELEMENT_TYPE_DIRECTORY) {
        if (strcmp(name, ".") == 0)
            return RT_SUCCESS;
        if (strcmp(name, "..") == 0)
            return RT_SUCCESS;

        unsigned int path_end_before = data->path_end;
        if (data->path_end > 0)
            data->path_scratch[data->path_end++] = '/';
        memcpy(data->path_scratch + data->path_end, name, name_len);
        data->path_scratch[data->path_end + name_len] = '\0';
        data->path_end += (unsigned int)name_len;

        rt_result res = rtIterateDirectory(data->path_scratch, user, DirectoryHandler);
        if (res != RT_SUCCESS)
            return res;

        data->path_end = path_end_before;
    }
    return RT_SUCCESS;
}

void rtDiscoverAssets(void) {

    rt_discovery_data data;
    memcpy(data.path_scratch, "assets", sizeof("assets"));
    data.path_end = sizeof("assets") - 1;
    
    rtIterateDirectory("assets", &data, DirectoryHandler);
}